import xml.etree.cElementTree as ElementTree
import logging

import sys
import re
import copy

import pydbm
import pycudd

logging.basicConfig()
logger = logging.getLogger('PYNTA')

class NTA:
  def __init__(self,templates=None):
    self.channels = []
    self.templates = templates or []
    
    #TODO: add global variables
    #self.variables = variables
    
  def add_template(self, t):
    if not t in self.templates:
      self.templates.append(t)
      
  def get_template(self,name):
    ts = [t for t in self.templates if t.name == name] 
    if len(ts) == 0:
      logger.error("No template with name \""+name+"\"")
      return None
    elif len(ts) > 1:
      logger.error("More than one template with name \""+name+"\"")
      return None
    else:
      return ts[0]

  #TODO: save to xml
  #def to_xml(self):
    #ret = """<?xml version="1.0" encoding="utf-8"?>
    #<!DOCTYPE nta PUBLIC "-//Uppaal Team//DTD Flat System 1.1//EN" "http://www.it.uu.se/research/group/darts/uppaal/flat-1_1.dtd">
    #<nta>
      #<declaration>""" + self.declaration + </declaration>
    
    #templatesxml = ""
    #for t in self.templates:
      #templatesxml += t.to_xml() + "\n"
      
    #ret += templatesxml
    #ret += "<system>" + self.system + "</system></nta>"
    #return ret
    
  def parse_declaration(self,declaration_string):
    ##TODO: parse other declarations type
    if declaration_string:
      decls = declaration_string.split('\n')   
      for decl in decls:
	if decl.startswith("broadcast chan "):
	  chan_names = decl[15:][:-1]
	  chan_names = chan_names.split(',')
	  for c in chan_names:
	    self.channels.append(c.strip())

  @classmethod
  def from_xml(cls, xmlsock):
    nta = cls()
    nta._from_xml(xmlsock)
    return nta

  
  def _from_xml(self, xmlsock):
    xmldoc = ElementTree.ElementTree(file=xmlsock).getroot()
    
    #TODO: add global clocks and global variables, system ? 
    declaration = xmldoc.findtext("declaration") or ""
    self.parse_declaration(declaration)
    
    #TODO: parse the position of the elements
  
    #TODO: add template parameters
    
    self.templates = []
    for templatexml in xmldoc.getiterator("template"):      
      locations = {}
      for locationxml in templatexml.getiterator("location"):
	name = locationxml.findtext("name") or ""
	id = locationxml.get('id')
        location = Location(name,id)
        
	if locationxml.find("committed") != None:
	  location.committed = True
	if locationxml.find("urgent") != None:
	  location.urgent = True
	                 
	for labelxml in locationxml.getiterator("label"):
	  if labelxml.get('kind') == 'invariant':
	    location.invariant = labelxml.text
	    
	locations[location.id] = location
      
      transitions = []
      for transitionxml in templatexml.getiterator("transition"):
	transition = Transition(locations[transitionxml.find('source').get('ref')],locations[transitionxml.find('target').get('ref')])
	transition.source.add_forward_transition(transition)
	transition.target.add_backward_transition(transition)

	transition.controllable = ('controllable', 'false') not in transitionxml.items() # parsed but changed according to synchronisation

        for labelxml in transitionxml.getiterator("label"):
	  if labelxml.get('kind') == 'assignment':
	    transition.update = labelxml.text
	  elif labelxml.get('kind') == 'guard':
	    transition.guard = labelxml.text  
	  elif labelxml.get('kind')  == 'synchronisation':
	    sync_type = labelxml.text[-1]
	    if sync_type == '?':
	      transition.controllable = False 	# this is not like in ECDAR
	    elif sync_type == '!':
	      transition.controllable = True
	    else:
	      logger.error("Unknown synchronisation type: \'"+sync_type+"\'")
	    channel_name = labelxml.text[:-1]
	    if channel_name in self.channels:
	      transition.synchronisation = (channel_name,sync_type)
	transitions.append(transition)
	  
      if templatexml.find("init") != None:
	init_location = locations[templatexml.find("init").get('ref')]
      else:
        init_location = None

      template = Template(templatexml.find("name").text,locations.values(),transitions,init_location)
      self.templates.append(template)
      
      template.parse_declaration(templatexml.findtext("declaration") or "")
      

class Template:
  def __init__(self,name,locations=None,transitions=None,init_location=None,variables=None,clocks=None,nb_fa=0,nb_fs=0,nb_fe=0):
    self.name = name
    self.locations = locations or {}
    self.transitions = transitions or {} 
    self.init_location = init_location

    self.variables = variables or {} # dictionary: links a variable name to an initial value
    self.clocks = clocks or {} # dictionary: links a clock name to the real time clock in the DBM context
    self.nb_fa = nb_fa
    self.nb_fs = nb_fs
    self.nb_fe = nb_fe
    
    self.context = None
    self.clock_bounds = {}
    self.mgr = None

    #TODO: add variables
    #self.variables = variables
    
  def get_location(self,loc_name):
    ls = [l for l in self.locations if l.name == loc_name] 
    if len(ls) == 0:
      logger.error("No location with name \""+loc_name+"\"")
      return None
    elif len(ls) > 1:
      logger.error("More than one location with name \""+loc_name+"\"")
      return None
    else:
      return ls[0] 
    
  def get_nb_features(self):
    return self.nb_fa + self.nb_fs + self.nb_fe
  
  def get_nb_adaptive_features(self):
    return self.nb_fa + self.nb_fe
    
  # TODO: print to xml
  #def to_xml(self):
    #ret =  "<template><name x=" + 5 + " y=" + 5 + ">" + self.name + "</name>" + "<declaration>"
    #for c in self.clocks
      #ret += "clock " + c + ";"
    #if self.fa:
      #ret += "bool fa[" + len(fa) + "] := {" + str(self.fa)[1:][:-1] + "} ;"
    #if self.fs:
      #ret += "bool fs[" + len(fs) + "] := {" + str(self.fs)[1:][:-1] + "} ;"
    #if self.fe:
      #ret += "bool fe[" + len(fe) + "] := {" + str(self.fe)[1:][:-1] + "} ;"
    #ret += "</declaration>"
    #for l in self.locations:
      #ret += l.to_xml() + "\n"
    #ret += "<init ref=" + self.init_location.id + " />\n"
    #for t in self.transitions:
      #ret += t.to_xml() + "\n"
    #ret += "</template>"
    #return ret
    
  def updateBound(self,clock, value):
    if self.clock_bounds[clock] is None:
      self.clock_bounds[clock] = value
    else:
      self.clock_bounds[clock] = max(self.clock_bounds[clock], value)
    
  def parse_declaration(self,declaration_string):
    if declaration_string:
      decls = declaration_string.split('\n')
      for decl in decls:
	decl = decl.strip()
	if not decl or decl.startswith("//"):
	  continue
	elif decl.startswith("clock "):
	  clock_names = decl[6:][:-1]
	  clock_names = clock_names.split(',')
	  for c in clock_names:
	    self.clocks[c.strip()] = None
	  #WARNING: no syntax check is performed. clocks may be declared two times, etc...
	  # we assume the model is checked in UPPAAL/ECDAR
	elif decl.startswith("bool "):
	  r = re.search('^bool\s+(\w+)\s*\[(\d+)\].*$', decl)
	  if r:
	    if r.group(1)=="Fa":
	      self.nb_fa = int(r.group(2))
	    elif r.group(1)=="Fs":
	      self.nb_fs = int(r.group(2))
	    elif r.group(1)=="Fe":
	      self.nb_fe = int(r.group(2))
	    else:
	      logger.warning("format not supported for boolean variable: "+r.group(1))
	  else:
	    logger.warning("format not supported for boolean variable: "+decl)
	elif decl.startswith("int "):	
	  r = re.search('^int\s+(\w+)\s*=\s*\d+\s*;$', decl)
	  if r:
	    self.variables[r.group(1)]=int(r.group(2))
	  else:
	    logger.warning("format not supported for integer variable: "+decl)
	else:
	  logger.warning("unknown variable type: "+decl)

  def parse_dbm_bdd(self):
    #TODO: reparsing clocks and bdd does not work
    #TODO: do not add the tatl_clock if not needed
    # create a DBM context with all the clocks plus a special tatl_clock
    self.context = pydbm.Context()
    self.clock_bounds = {}
    SetOfSymbolicStates.context = self.context
    for clock_name in self.clocks:
      context_clock_name = self.name+'.'+clock_name
      self.context.addClockByName(context_clock_name)
      clock = self.context[context_clock_name]
      self.clocks[clock_name] = clock
      self.clock_bounds[clock] = None
    self.context.addClockByName("tatl_clock")
    clock = self.context["tatl_clock"]
    self.clocks["tatl_clock"] = clock
    self.clock_bounds[clock] = None
    
    # create a BDD manager
    self.mgr = pycudd.DdManager(self.get_nb_features() + self.get_nb_adaptive_features())
    self.mgr.SetDefault()
    SetOfSymbolicStates.mgr = self.mgr

    # parse the guards and invariants into DBMs and BDDs
    for location in self.locations:
      location.parse_invariant(self)
      
    for transition in self.transitions:
      transition.parse_guard(self)
      transition.parse_update(self)
	    
  def parse_dbm(self, constraints):
    #print "parse dbm: " + constraints + "\n"

    if not constraints:
      return None
      
    constraints = constraints.replace('&&','and')
    constraints = constraints.split('and')
    constraints_dbm = []
    
    for c in constraints:
      if c.strip():
	constraints_dbm.append(c)
	r = re.search('^\s*(?P<clock>\w+)\s*((>=)|(>)|(==)|(<=)|(<))\s*(?P<coeff>\d+)\s*$',c)
	if r is None:
	  r = re.search('^\s*(?P<coeff>\d+)\s*((>=)|(>)|(==)|(<=)|(<))\s*(?P<clock>\w+)\s*$',c)
	if r is None:
	  logger.error('Cannot parse constraint ' + c)
	  return None
	if self.clocks.has_key(r.group("clock")):
	  clock = self.clocks[r.group("clock")]
	  value = int(r.group('coeff'))
	  self.updateBound(clock,value)
	else:
	  logger.error('Unknown clock \''+r.group("clock")+'\'')
	  return None
	  
    constraints_dbm = '&'.join(['(' + c + ')' for c in constraints_dbm])
    code = compile(constraints_dbm, "<dbm>", 'eval') 
    try:
      dbm = eval(code, self.clocks)            
      assert(isinstance(dbm, pydbm.Federation))
      #print str(dbm)
    except Exception as e:
      logger.error('Cannot parse dbm from constraints "' + constraints_dbm + '": ' + str(e))
      dbm = None
    return dbm
  
  def parse_bdd(self, expr):
    #print "parse bdd: " + expr + "\n"
    
    if not expr:
      return None
      
    expr = expr.replace('&&','&')
    expr = expr.replace('and','&')
    expr = expr.replace('||','|')
    expr = expr.replace('or','|')
    expr = expr.replace('!','~')
    expr = expr.replace('not','~')
    expr = expr.replace("[","")
    expr = expr.replace("]","")
    
    features = {}
    for index in range(0,self.get_nb_features() + self.get_nb_adaptive_features()):
      features[self.get_feature_name(index)] = self.mgr.IthVar(index)
    
    #for num in range(0,self.nb_fa):
      #features["Fa"+str(num)] = self.mgr.IthVar(num)
    
    #for num in range(0,self.nb_fs):
      #features["Fs"+str(num)] = self.mgr.IthVar(num+self.nb_fa)

    #for num in range(0,self.nb_fe):
      #features["Fe"+str(num)] = self.mgr.IthVar(num+self.nb_fa+self.nb_fs)
    
    code = compile(expr, "<bdd>", 'eval') 
    try:
      bdd = eval(code, features)            
      assert(isinstance(bdd, pycudd.DdNode))
      #print self.bdd_to_str(bdd)
    except Exception as e:
      logger.error('Cannot parse bdd from expression "' + expr + '": ' + str(e))
      bdd = None      
    return bdd
      
  def bdd_to_str(self,bdd):
    #print "*** print bdd to str ***"
    if bdd.IsConstant():
      if bdd.Empty():
	return "false"
      else:
	return "true"
    
    ret=""
    pycudd.set_iter_meth(0)
    for prime in bdd:
      if ret:
	ret += " or "
      ret += "("
      index = 0
      printed = False
      for p in prime:
	if p != 2:
	  if printed:
	    ret += " and "
	  if p == 0:
	    ret += "~" + self.get_feature_name(index)
	  elif p == 1:
	    ret += self.get_feature_name(index)
	  printed = 1
  	index +=1
      ret += ")"
    return ret
  
  def symbolic_state_to_str(self,state):
    return "("+  self.bdd_to_str(state.bdd) + ", " + str(state.zone) + ")"
    
  def set_of_symbolic_state_to_str(self,states):
    ret = ""
    for state in states:
      if ret:
	ret += ", "
      ret += self.symbolic_state_to_str(state)
    return "[" + ret + "]"  

  def get_feature_name(self,index):
    if index < self.nb_fs:
      return "Fs" + str(index) + ""
 
    index -= self.nb_fs
    if index < self.nb_fa:
      return "Fa" + str(index) + ""
    
    index -= self.nb_fa   
    if index < self.nb_fe:
      return "Fe" + str(index) + ""
    
    index -= self.nb_fe
    if index < self.nb_fa:
      return "Fb" + str(index) + ""
    
    index -= self.nb_fa
    if index < self.nb_fe:
      return "Ff" + str(index) + ""
   
    return ""
  
  def get_feature_variable(self,index):
    if index < self.get_nb_features() + self.get_nb_adaptive_features():
      return self.mgr.IthVar(index)
    else:
      logger.error("Index \'" + str(index) + " is out of features range")
      sys.exit(1)
      
      
class Location:
  def __init__(self,name,id,commited=False,urgent=False,invariant=""):
    self.name = name
    self.id = id
    self.commited = commited
    self.urgent = urgent
    self.invariant = invariant
    self.invariant_dbm = None
    self.invariant_bdd = None
    self.forward = []
    self.backward = []
    
  def set_position(self,posx,posy):
    self.posx = posx
    self.posy = posy
    
  def add_forward_transition(self,t):
    self.forward.append(t)
        
  def add_backward_transition(self,t):
    self.backward.append(t)
    
  def parse_invariant(self, template):
    feature_inv = self.invariant[self.invariant.find('(')+1:]
    clock_inv = self.invariant
    if feature_inv == self.invariant:
      feature_inv = ""
    else:  
      feature_inv = feature_inv[:-1]
      clock_inv = self.invariant[:self.invariant.find('(')]
    
    self.invariant_dbm = template.parse_dbm(clock_inv) or template.context.getTautologyFederation()
    self.invariant_bdd = template.parse_bdd(feature_inv) or template.mgr.One()

    
    
    
  # TODO  
  #def to_xml(self):
    #ret = "<location id=" + "
        #namexml = self.name.to_xml()
        #invariantxml = self.invariant.to_xml()
        #return """
    #<location id="%s" x="%s" y="%s">
      #%s
      #%s
      #%s
      #%s
    #</location>""" % (self.id, self.xpos, self.ypos, namexml, invariantxml,
        #self.committed and '<committed />' or '', self.urgent and '<urgent />' or '')
    

class Transition:
  def __init__(self,source,target,select=None,guard="",synchronisation=None,update="",controllable=True):
    self.source = source
    self.target = target
    self.select = select
    self.guard = guard
    self.guard_dbm = None
    self.guard_bdd = None
    self.synchronisation = synchronisation
    self.update = update
    self.update_clock = []
    self.adaptive_features = []   # all the adaptive features minus the one fixed in the update
    self.controllable = controllable
    
  def parse_guard(self, template):    
    feature_guard = self.guard[self.guard.find('(')+1:]
    clock_guard = self.guard
    if feature_guard == self.guard:
      feature_guard = ""
    else:
      feature_guard = feature_guard[:-1]
      clock_guard = self.guard[:self.guard.find('(')]
    
    self.guard_dbm = template.parse_dbm(clock_guard) or template.context.getTautologyFederation()
    self.guard_bdd = template.parse_bdd(feature_guard) or template.mgr.One()
    
  def parse_update(self,template):
    #print "*** parsing updates ***"
    #TODO: check if the adaptive features not updated are not changed in the guard
    if self.controllable: # system transition, not like ECDAR
      for i in range(0,template.nb_fa):
	self.adaptive_features.append(i+template.nb_fs)
    else: # environment transition
      for i in range(0,template.nb_fe):
	self.adaptive_features.append(i+template.nb_fs+template.nb_fa)
    
    #TODO: extend syntax of assignments
    if self.update:
      for u in self.update.split(','):
	#print "update: " + u
	r = re.search('^\s*(\w+)\s*=\s*(\d+)', u)
	if r:
	  varname = r.group(1)
	  if template.clocks.has_key(varname):
	    clock = template.clocks[varname]
	    value = int(r.group(2))
	    if value != 0:
	      logger.error("Clock reset to non-zero value")
	    else:
	      self.update_clock.append((clock,value))
	      template.updateBound(clock,value)
	  else:
	    logger.error('Unknown variable ' + varname + ' in assignment ' + u)
	  continue
	
	r = re.search('^\s*(\w+)\[(\d+)\]\s*=\s*(\w+)\[(\d+)\]\s*', u)
	if r:
	  varname1 = r.group(1)
	  index1 = int(r.group(2))
	  varname2 = r.group(3)
	  index2 = int(r.group(4))
	  if self.controllable and ((varname1 == "Fa" and varname2 == "Fb") or (varname1 == "Fb" and varname2 == "Fa")):
	    assert(index1 == index2)
	    self.adaptive_features.remove(index1+template.nb_fs)
	  elif not self.controllable and ((varname1 == "Fe" and varname2 == "Ff") or (varname1 == "Ff" and varname2 == "Fe")):
	    assert(index1 == index2)
	    self.adaptive_features.remove(index1+template.nb_fs+template.nb_fa)
	  else:
	    logger.error('Syntax error in assignment ' + u)
	    #TODO: output a better error message
	else:
	  logger.error('Cannot parse assignment ' + u)
	  
  def get_guard_bdd(self,template):
    #print "Guard bdd" 
    # return the bdd that only condition the needed features
    ret = copy.copy(self.guard_bdd)
    for f in self.adaptive_features:
      ret = ret.ExistAbstract(template.get_feature_variable(f+template.get_nb_adaptive_features()))
    #print template.bdd_to_str(ret)
    return ret
  
  # TODO  
  #def to_xml(self):
  
  
class SymbolicState:
  def __init__(self,bdd,zone):
    self.bdd = bdd
    self.zone = zone
  
  def copy(self):
    ret = copy.copy(self)
    ret.bdd = copy.copy(self.bdd)
    ret.zone = self.zone.copy()         
    return ret
  
class SetOfSymbolicStates(list):  
  context = None
  mgr = None
  
  def __init__(self):
    pass
  
  def copy(self):
    ret = self.__class__()
    for state in self:
      ret += [state.copy()]
    return ret 
       
  def append(self,new_state):
    """ Add a new symbolic state.
	Apply reduction rules to merge with the states already in.
	Modify new_state and return the part of the symbolic state that is new in the list.
    """
    if (new_state.bdd.Empty() | new_state.zone.isEmpty()):
      return None
      
    index = 0
    equal_state = None
    for state in self:
      if new_state.zone <= state.zone:
	#print "SMALLER"
	new_state.bdd = new_state.bdd & ~state.bdd
	if new_state.bdd.Empty():
	  return None
	if new_state.zone == state.zone:
	  #print "EQUAL"
	  #assert(not equal_state)  #only one zone may be equal because otherwise the two already present should have been merged
	  equal_state = state
      elif state.zone < new_state.zone:
	#print "BIGGER"
	state.bdd = state.bdd & ~new_state.bdd
	if state.bdd.Empty():
	  self.pop(index)
      index += 1
    if equal_state:
      #new_state.bdd = new_state.bdd & ~equal_state.bdd # already done 14 lines above
      equal_state.bdd = equal_state.bdd | new_state.bdd
      if new_state.bdd.Empty():
	return None
    else:
      self += [new_state]
    return new_state
  
  def __invert__(self):    
    result = SetOfSymbolicStates()
    if not self:
      result += [SymbolicState(SetOfSymbolicStates.mgr.One(),SetOfSymbolicStates.context.getTautologyFederation())]
      return result
    
    current_state = self[0]
    result.append(SymbolicState(current_state.bdd,~current_state.zone))
    result.append(SymbolicState(~current_state.bdd,SetOfSymbolicStates.context.getTautologyFederation()))
    
    for index in range(1,len(self)):
      current_state = self[index]
      old_results = result
      result = SetOfSymbolicStates()
      for state in old_results:
	result.append(SymbolicState(state.bdd & current_state.bdd,state.zone & ~current_state.zone))
	result.append(SymbolicState(state.bdd & ~current_state.bdd,state.zone))
    return result
  
  def __or__(self,arg):
    result = self.copy()
    for state in arg:
      result.append(state.copy())
    return result
  
  def __ior__(self,arg):
    for state in arg:
      self.append(state.copy())
    return self
    
  def __and__(self,arg):
    result = SetOfSymbolicStates()
    for state_self in self:
      for state_arg in arg:
	result.append(SymbolicState(state_self.bdd & state_arg.bdd, state_self.zone & state_arg.zone))
    return result
  
  def __iand__(self,arg):
    result = SetOfSymbolicStates()
    for state_self in self:
      for state_arg in arg:
	result.append(SymbolicState(state_self.bdd & state_arg.bdd, state_self.zone & state_arg.zone))
    self = result
    return self
  
  def __sub__(self, arg):
    result = self.copy()
    result &= ~arg
    return result
  
  def __isub__(self, arg):
    self &= ~arg
    return self

  def conjunction(self,state):
    result = SetOfSymbolicStates()
    for current_state in self:
      result.append(SymbolicState(current_state.bdd & state.bdd, current_state.zone & state.zone))
    return result
  
  def conjunction_in_place(self,state):
    for index in xrange(len(self) - 1, -1, -1):
      current_state = self[index]
      current_state.bdd &= state.bdd
      current_state.zone &= state.zone
      if (current_state.bdd.Empty() | current_state.zone.isEmpty()):
	del self[index]
    return self
    
  def difference(self,big_state):
    neg = ~self
    neg.conjunction_in_place(big_state)
    return neg
 
  def freeClockInPlace(self,clock):
    for state in self:    
      state.zone.freeClockInPlace(clock)
    return self
  
  def pred(self, template, transition):
    #print "computing predecessors of " + template.set_of_symbolic_state_to_str(self)

    # computes the predecessors of this through transition
    ret_states = SetOfSymbolicStates()
    for state in self:    
      #print "computing predecessors of state " + template.symbolic_state_to_str(state)

      ret_state = state.copy()
      ret_state.bdd &= transition.target.invariant_bdd
      ret_state.zone &= transition.target.invariant_dbm
      
      # compute the preimage of the zone
      for (clock, value) in transition.update_clock:
	ret_state.zone &= (clock == 0)
	ret_state.zone.freeClockInPlace(clock)
      ret_state.zone &= transition.guard_dbm
      ret_state.zone &= transition.source.invariant_dbm

      #print "zone predecessors : " + str(ret_state.zone)

      # compute the preimage of the bdd
      nb_adaptive_features = len(transition.adaptive_features)
      swap_old = pycudd.DdArray(nb_adaptive_features)
      swap_new = pycudd.DdArray(nb_adaptive_features)
      for f in transition.adaptive_features:
	swap_old.Push(template.get_feature_variable(f))
	swap_new.Push(template.get_feature_variable(f+template.get_nb_adaptive_features()))
      ret_state.bdd = ret_state.bdd.SwapVariables(swap_old,swap_new,nb_adaptive_features)
      #print "swap variables : " + template.bdd_to_str(ret_state.bdd)
      ret_state.bdd &= transition.guard_bdd
      #print "add guard : " + template.bdd_to_str(ret_state.bdd)
      for f in transition.adaptive_features:  
	ret_state.bdd = ret_state.bdd.ExistAbstract(template.get_feature_variable(f+template.get_nb_adaptive_features()))
	#print "abstract adaptive feature " + str(f+template.get_nb_adaptive_features()) + " : " + template.bdd_to_str(ret_state.bdd)
      ret_state.bdd &= transition.source.invariant_bdd
      #print "bdd predecessors : " + template.bdd_to_str(ret_state.bdd)

      ret_states.append(ret_state)
    return ret_states
    
  def predt(self, template, bad_states):
    # computes the time predecessors of this that avoid bad_states
    ret_states = SetOfSymbolicStates()
    for good_state in self:
      pred_good_zone = good_state.zone.predt(template.context.getEmptyFederation())
      current_preds = SetOfSymbolicStates()
      current_preds.append(SymbolicState(good_state.bdd, good_state.zone.predt(template.context.getEmptyFederation())))
      
      for bad_state in bad_states:
	bad_state_preds = SetOfSymbolicStates()
	bad_state_pred1 = SymbolicState(~bad_state.bdd,template.context.getTautologyFederation())
	if not bad_state_pred1.bdd.Empty():
	  bad_state_preds += [bad_state_pred1]
	bad_state_pred2 = SymbolicState(bad_state.bdd, good_state.zone.predt(bad_state.zone))
	if not bad_state_pred2.bdd.Empty() and not bad_state_pred2.zone.isEmpty():
	  bad_state_preds += [bad_state_pred2]
	current_preds &= bad_state_preds
      ret_states |= current_preds
    return ret_states
    
    
  