# TODO: create simple safety games and initialize location.accepting

import sys
import logging
import copy

import pynta
import pytatl
import pyparamexpr
import pyppl
import pycudd
    
logging.basicConfig()
logger = logging.getLogger('TIGA')
   
class TimedGame:
    def __init__(self,goal_template,objective):
	"""Create a game on the goal template whose objective is:
	- A[phi Until psi] if game_type is true (reachability type: e.g. A[True Until psi]=A<> psi)
	- A[phi WeakUntil psi] if game_type is false (safety type: e.g. A[not phi WeakUntil False]=A[] not phi)
	phi and psi are set of states (dictionaries) in the goal template: each one associate a goal location to a goal zone. 
	"""
        self.goal_template = goal_template
        if type(objective) == pytatl.TATLFormula:
	  self.objective = objective
        else:
	  self.objective = pytatl.TATLFormula(self.goal_template,objective)
        
        self.context = self.goal_template.context
        self.mgr = self.goal_template.mgr
        self.results = None

    def init_game():
        for l in self.goal_template.locations:
	  l.zone = self.context.getEmptyFederation()
	  l.depend = []
	  l.accepting_zone = self.context.getEmptyFederation()
        self._forward = {} 	# dictionnary that store for each state the next reachable states
        self.strategy = []	# ordered list of the states refined
	self.initial_loc = self.goal_template.initlocation
	initial_zone = self.context.getZeroFederation()
        initial_zone.upInPlace()
        initial_zone &= initial.invariant
        initial_zone.extrapolateMaxBoundsInPlace(self.model.clock_bounds)
        self.waiting = [DiscreteTransition(None,None,initial_zone,self.initial_loc,None)] # specialgame transition

    def solve(self):
	self.results = self._solve(self.objective.formula)
	return self.get_winning_features()
    
    def get_winning_features(self):
	bdd = self.mgr.Zero()
	for state in self.results[self.goal_template.init_location]:
	  if state.zone.hasZero():
	    bdd |= state.bdd
	return bdd
      
    def get_deadlock_state(self,location,controllable):
	# get the states beyond the invariant (so in deadlock) because one player didn't due the action he was forced to do
	bound_invariant = location.invariant_dbm & ~location.invariant_dbm.tightenUp()
	deadlocks = pynta.SetOfSymbolicStates()
        for t in location.forward:
	  if t.controllable == controllable:
	    guard_inv_bdd = t.get_guard_bdd(self.goal_template) & location.invariant_bdd
	    guard_inv_dbm = t.guard_dbm & bound_invariant
	    if not guard_inv_bdd.Empty() and not guard_inv_dbm.isEmpty():
	      deadlocks.append(pynta.SymbolicState(guard_inv_bdd,guard_inv_dbm.up()-guard_inv_dbm))
	  #print "BDD non deadlock:"
	  #print self.goal_template.bdd_to_str(bdd_not_deadlock)
	return deadlocks

    def _solve(self,formula):	
      	if type(formula) == dict:
	  assert(len(formula) == len(self.goal_template.locations))
	  return formula
	
      	if len(formula) == 0:
	  logger.error("No formula provided.")
	  return {}
	
	if len(formula) == 1:
	  return self._solve(formula[0])
  
	if len(formula) == 2:
	  if type(formula[0]) == pycudd.DdNode:
	    formula[1] = self._solve(formula[1])
	    return self.solve_feature(formula[0],formula[1])
	  if type(formula[0]) == str and formula[0] == "not":
	    formula[1] = self._solve(formula[1])
	    return self.solve_not(formula[1])

	if len(formula) == 3:
	  formula[0] = self._solve(formula[0])
	  formula[2] = self._solve(formula[2])  
	  if type(formula[1]) == str and formula[1] == "and":
	    return self.solve_and(formula[0],formula[2])
	  if type(formula[1]) == str and  formula[1] == "or":
	    return self.solve_or(formula[0],formula[2])
	    
	if len(formula) == 4:
  	  formula[1] = self._solve(formula[1])
	  formula[3] = self._solve(formula[3])  
	  if type(formula[0]) == str and formula[0] == "A":    
	    if type(formula[2]) == str and formula[2] == "U":
	      return self.solve_A_until(formula[1],formula[3])
	    if type(formula[2]) == str and formula[2] == "W":
	      return self.solve_A_weakuntil(formula[1],formula[3])
	  if type(formula[0]) == str and formula[0] == "E":    
	    if type(formula[2]) == str and formula[2] == "U":
	      return self.solve_E_until(formula[1],formula[3])
	    if type(formula[2]) == str and formula[2] == "W":
	      return self.solve_E_weakuntil(formula[1],formula[3])
	    
	if len(formula) == 5:
  	  formula[1] = self._solve(formula[1])
	  formula[4] = self._solve(formula[4])
	  minval = formula[3][0]
	  maxval = formula[3][1]
	  if type(formula[0]) == str and formula[0] == "A":    
	    if type(formula[2]) == str and formula[2] == "U":
	      return self.solve_A_until(formula[1],formula[4],minval,maxval)
	  if type(formula[0]) == str and formula[0] == "E":    
	    if type(formula[0]) == str and formula[2] == "U":
	      return self.solve_E_until(formula[1],formula[4],minval,maxval)
	    
	logger.error("Unknown formula: " + str(formula))
	return {}
      
    def init_tatl_clock(self,minval,maxval,psi):
      	assert(minval != "0" or maxval != "inf")
	tatl_clock = self.context["tatl_clock"]
	minval = int(minval)
	tatl_dbm = tatl_clock >= minval
	if maxval != "inf":
	  maxval = int(maxval)
	  self.goal_template.updateBound(tatl_clock,maxval)
	  tatl_dbm &= tatl_clock <= maxval
	
	for loc in self.goal_template.locations:
	  if psi[loc]:
	    psi[loc].conjunction_in_place(pynta.SymbolicState(self.mgr.One(),tatl_dbm))
	return tatl_clock
      
    def solve_not(self,phi):
	#print "*** Solving not ***"
	result = {}
	for l in self.goal_template.locations:
	  result[l] = ~phi[l]
	  for state in result[l]:
	    state.bdd &= l.invariant_bdd
	    state.zone &= l.invariant_dbm
	return result
    
    def solve_feature(self,bdd,phi):
	#print "*** Solving feature ***"
      	result = {}
	for l in self.goal_template.locations:
	  result[l] = pynta.SetOfSymbolicStates()
	  for state in phi[l]:
	    result[l].append(pynta.SymbolicState(bdd & state.bdd, state.zone))
	  result[l].append(pynta.SymbolicState(~bdd & l.invariant_bdd, l.invariant_dbm.copy()))
	return result
	  
    def solve_or(self,phi,psi):
	#print "*** Solving or ***"
	result = {}
	for l in self.goal_template.locations:
	  result[l] = phi[l] | psi[l]
	return result
    
    def solve_and(self,phi,psi):
	#print "*** Solving and ***"
      	result = {}
	for l in self.goal_template.locations:
	  result[l] = phi[l] & psi[l]
	return result

    def solve_A_until(self,phi,psi,minval="0",maxval="inf"):
	#print "*** Solving A( )U( ) ***"
	#TODO: test if the objective is limited to the reachability of some locations.
	# then use a reachability game
	tatl_clock = None
	if minval != "0" or maxval != "inf":
	  tatl_clock = self.init_tatl_clock(minval,maxval,psi)
	  
	winning = {}
	waiting = []
	for loc in self.goal_template.locations:
	  # we assume psi[loc] is included in the location invariant
	  winning[loc] = psi[loc].copy()
	  if winning[loc]:
	    waiting.append(loc)
	
	new_states = True
        while waiting:
	  loc = waiting.pop(0)
	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          good = winning[loc].copy()
          good |= self.get_deadlock_state(loc,False) & ~self.get_deadlock_state(loc,True) # the system wins if the environnement must do a transition to avoid a deadlock and the system can't
	  for t in loc.forward:
	    if t.controllable:
	      good |= winning[t.target].pred(self.goal_template,t)
  	  #print "good states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(good)

	  bad = ~(phi[loc]) & ~(winning[loc])
	  bad.conjunction_in_place(loc_inv)
	  for t in loc.forward:
	    if not t.controllable:
	      bad |= (~winning[t.target]).pred(self.goal_template,t)
       	  #print "bad states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(bad)

	  new_winning = good.predt(self.goal_template,bad)
  	  new_winning.conjunction_in_place(loc_inv)
    	  #print "new_winning states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(new_winning)
	  #print "old winning states in location \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(winning[loc])

	  for state in new_winning:
	    #print "new_winning state in location \'" + loc.name + "\' : " + self.goal_template.symbolic_state_to_str(state)
	    if winning[loc].append(state):
	      new_states = True
	  #print "winning states sin location \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(winning[loc])	  
	  if new_states:
	    for t in loc.backward:
	      if t.source not in waiting:
		waiting.append(t.source)
	    new_states = False
	
	if tatl_clock:
	  for loc in self.goal_template.locations:
	    winning[loc].conjunction_in_place(pynta.SymbolicState(self.mgr.One(),tatl_clock == 0))
	    winning[loc].freeClockInPlace(tatl_clock)
	return winning
	      
    def solve_A_weakuntil(self,phi,psi):
	#print "*** Solving A( )W( ) ***"
	#TODO: test if the objective is limited to the reachability of some locations.
	# then use a reachability game
	losing = {}
	waiting = []
	for loc in self.goal_template.locations:
  	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
  	  losing[loc] = ~psi[loc] & ~phi[loc]
  	  losing[loc].conjunction_in_place(loc_inv)
	  if losing[loc]:
	    waiting.append(loc)

	new_states = True
        while waiting:
	  loc = waiting.pop(0)
	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          bad = losing[loc].copy()
       	  #print "bad states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(bad)
          bad |= self.get_deadlock_state(loc,True) # the environnement wins if the system can do a transition to avoid a deadlock
	  for t in loc.forward:
	    if not t.controllable:
	      bad |= losing[t.target].pred(self.goal_template,t)
  	  #print "bad states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(bad)

	  good = psi[loc].copy()
	  for t in loc.forward:
	    if t.controllable:
	      good |= (~losing[t.target]).pred(self.goal_template,t)
       	  #print "good states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(good)

	  new_losing = bad.predt(self.goal_template,good-bad)
  	  new_losing.conjunction_in_place(loc_inv)
    	  #print "new losing states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(new_losing)

	  for state in new_losing:
	    #print "append state in location \'" + loc.name + "\' : " + self.goal_template.symbolic_state_to_str(state)
	    if losing[loc].append(state):
	      #print "new losing state in location \'" + loc.name + "\' : " + self.goal_template.symbolic_state_to_str(state)
	      new_states = True
	  #print "losing states in location \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(losing[loc])
	  if new_states:
	    for t in loc.backward:
	      if t.source not in waiting:
		waiting.append(t.source)
	    new_states = False
	
	winning = {}
	for loc in self.goal_template.locations:
	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
	  winning[loc] = (~losing[loc]).conjunction_in_place(loc_inv)
	return winning

    def solve_E_until(self,phi,psi,minval="0",maxval="inf"):
	#print "*** Solving E( )U( ) ***"
	#TODO: test if the objective is limited to the reachability of some locations.
	# then use a reachability game
	tatl_clock = None
	if minval != "0" or maxval != "inf":
	  tatl_clock = self.init_tatl_clock(minval,maxval,psi)
	
	winning = {}
	waiting = []
	for loc in self.goal_template.locations:
	  # we assume psi[loc] is included in the location invariant
	  winning[loc] = psi[loc].copy()
	  if winning[loc]:
	    waiting.append(loc)

	new_states = True
        while waiting:
	  loc = waiting.pop(0)
	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          good = winning[loc].copy()
	  for t in loc.forward:
	    good |= winning[t.target].pred(self.goal_template,t)
  	  #print "good states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(good)

	  bad = ~(phi[loc]) & ~(winning[loc])
	  bad.conjunction_in_place(loc_inv)
       	  #print "bad states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(bad)

	  new_winning = good.predt(self.goal_template,bad)
  	  new_winning.conjunction_in_place(loc_inv)
    	  #print "new_winning states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(new_winning)
	  #print "old winning states in location \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(winning[loc])

	  for state in new_winning:
	    #print "new_winning state in location \'" + loc.name + "\' : " + self.goal_template.symbolic_state_to_str(state)
	    if winning[loc].append(state):
	      new_states = True
	  #print "winning states sin location \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(winning[loc])	  
	  if new_states:
	    for t in loc.backward:
	      if t.source not in waiting:
		waiting.append(t.source)
	    new_states = False
	
	if tatl_clock:
	  for loc in self.goal_template.locations:
	    winning[loc].conjunction_in_place(pynta.SymbolicState(self.mgr.One(),tatl_clock == 0))
    	    winning[loc].freeClockInPlace(tatl_clock)
	return winning
	      
    def solve_E_weakuntil(self,phi,psi):
	#print "*** Solving E( )W( ) ***"
	#TODO: test if the objective is limited to the reachability of some locations.
	# then use a reachability game
	losing = {}
	waiting = []
	for loc in self.goal_template.locations:
  	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
  	  losing[loc] = ~psi[loc] & ~phi[loc]
  	  losing[loc].conjunction_in_place(loc_inv)
	  if losing[loc]:
	    waiting.append(loc)
	
	new_states = True
        while waiting:
	  loc = waiting.pop(0)
	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
          
          bad = losing[loc].copy()
          bad |= self.get_deadlock_state(loc,False) | self.get_deadlock_state(loc,True) # the system loses if either the environnement or the system must do a transition to avoid a deadlock

	  good = psi[loc].copy()
	  for t in loc.forward:
	    good |= (~losing[t.target]).pred(self.goal_template,t)
       	  #print "good states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(good)

	  new_losing = bad.predt(self.goal_template,good-bad)
  	  new_losing.conjunction_in_place(loc_inv)
    	  #print "new losing states in \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(new_losing)

	  for state in new_losing:
	    if losing[loc].append(state):
	      #print "new losing state in location \'" + loc.name + "\' : " + self.goal_template.symbolic_state_to_str(state)
	      new_states = True
	  #print "losing states in location \'" + loc.name + "\' : " + self.goal_template.set_of_symbolic_state_to_str(losing[loc])
	  if new_states:
	    for t in loc.backward:
	      if t.source not in waiting:
		waiting.append(t.source)
	    new_states = False
	
	winning = {}
	for loc in self.goal_template.locations:
	  loc_inv = pynta.SymbolicState(loc.invariant_bdd,loc.invariant_dbm)
	  winning[loc] = (~losing[loc]).conjunction_in_place(loc_inv)
	return winning
      
      
    #def solve_A_until(self,phi,psi,all_state):
	## all_state: compute and return all the winning states;
	## not all_state: stopped when the initial state is winning; return True or False.
	##TODO: test if the objective is limited to the reachability of some locations.
	## then use a reachability game
      
	#self.init_game()
	#solved = False
        #i = 0
        #while self.waiting:
            #e = self.waiting.pop()
            #t = e.model_transition
            #S = e.source_zone
            #loc = e.source_loc
            #S_ = e.target_zone
            #loc_ = e.target_loc
            #i += 1
            #if not S_ <= loc_.zone: # not included
	      #logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      #loc_.zone |= S_
	      #for e_ in self.next(loc_):
		#logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		#if not e_ in  self.waiting: # Can the contrary happen ?
		  #self.waiting.append(e_)
		#else:
		  #logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      #loc_.accepting_zone = loc_.zone & psi[loc_]
	      #if loc_.accepting_zone != loc_.zone:
		#bad = loc_.zone - phi[loc_]
		#for a in self.nextUncontrollable(loc_):
		  #bad  |= a.pred(a.target_zone - a.target_loc.accepting_zone)
		#loc_.accepting_zone = loc_.accepting_zone.predt(bad)

	      #if not (S_ & loc_.accepting_zone).isEmpty():
		#logger.debug("adding to wait accepting: " + str(loc_))
		#self.waiting.append(e)
            #else:                
	      #logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      #good = loc.accepting_zone
	      #for a in self.nextControllable(loc):
		#good |= a.pred(a.target_loc.accepting_zone)
	      #bad = loc.zone - phi[loc]
	      #for a in self.nextUncontrollable(loc):
		#bad  |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      #accepting_ = good.predt(bad)
	      #accepting_ &= loc.zone
	      ##print accepting_

	      #if accepting_ != loc.accepting_zone:
		#self.strategy.append(loc)
		#for a in loc.depend:
		  #logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  #e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  #if not e_ in  self.waiting:
		     #self.waiting.append(e_)
		#loc.accepting_zone = accepting_
		#logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      ##loc_.depend.add(t)  
	      #if not t in loc_.depend:
		#loc_.depend.append(t)
	      #logger.debug('Accepting zone:' + str(loc.accepting_zone))
	    
	    #if  self.initial_loc.accepting_zone.hasZero():
	      #solved = True
              #if not all_state: 
		#break

	#logger.info(str(i) + " states visited.")
	
	#if all_state: 
	  #result = {}
	  #for l in self.goal_template.locations:
	    #result[l] = l.accepting_zone
	  #return result
	#else:
	  #return solved

    #def solve_A_weakuntil(self,phi,psi,all_state):
	## all_state: compute and return all the winning states;
	## not all_state: stopped when the initial state is losing; return True or False.
	##TODO: test if the objective is limited to a safety condition w.r.t. some locations.
	## then use a safety game
	
	#init_game()
	#solved = True
	#i = 0
        #while self.waiting:
            #e = self.waiting.pop()
            #t = e.model_transition
            #S = e.source_zone
            #loc = e.source_loc
            #S_ = e.target_zone
            #loc_ = e.target_loc
            #i += 1
            #if not S_ <= loc_.zone: # not included
	      #logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      #loc_.zone |= S_
	      #for e_ in self.next(loc_):
		#logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		#if not e_ in self.waiting: # Can the contrary happen ?
		    #self.waiting.append(e_)
		#else:
		    #logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      #loc_.accepting_zone = loc_.zone - psi[loc_] - phi[loc_]
	      #if loc_.accepting_zone != loc_.zone:
		#good  = psi[loc_]
		#for a in self.nextControllable(loc_):
		  #good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
		#loc_.accepting_zone = loc_.accepting_zone.predt(good-loc_.accepting_zone)
	      
	      #if not (S_ & loc_.accepting_zone).isEmpty():
		#logger.debug("adding to wait accepting: " + str(loc_))
		#self.waiting.append(e)
            #else:                
	      #logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      #bad = loc.accepting_zone
	      #bad += loc.zone.up() - loc.zone # lm comment: S.zone.up() - S.zone is the zone beyond the invariant. In this zone the environnement is winning.
	      #for a in self.nextUncontrollable(loc):
		  #bad  |= a.pred(a.target_loc.accepting_zone)
	      ##print "bad: " + str(bad)
	      #good  = psi[loc]
	      #for a in self.nextControllable(loc):
		  #good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      #accepting_ = bad.predt(good-bad)
	      #accepting_ &= loc.zone
	      ##print accepting_

	      #if accepting_ != loc.accepting_zone:
		#self.strategy.append(loc)
		#for a in loc.depend:
		  #logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  #e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  #if not e_ in self.waiting:
		    #self.waiting.append(e_)
		#loc.accepting_zone = accepting_
		#logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      ##loc_.depend.add(t)  
	      #if not t in loc_.depend:
		#loc_.depend.append(t)
	      #logger.debug('Accepting zone:' + str(loc.accepting_zone))
	          
	    #if self.initial_loc.accepting_zone.hasZero():
              #self.solved = False
	      #if not all_state: 
		#break

	#logger.info(str(i) + " states visited.")
		
	#if all_state: 
	  #result = {}
	  #for l in self.goal_template.locations:
	    #result[l] = l.zone - l.accepting_zone
	  #return result
	#else:
	  #return solved
    
    #def solve_E_until(self,phi,psi,all_state):
	## all_state: compute and return all the winning states;
	## not all_state: stopped when the initial state is winning; return True or False.
	##TODO: test if the objective is limited to the reachability of some locations.
	## then use a reachability game
      
	#self.init_game()
	#solved = False
        #i = 0
        #while self.waiting:
            #e = self.waiting.pop()
            #t = e.model_transition
            #S = e.source_zone
            #loc = e.source_loc
            #S_ = e.target_zone
            #loc_ = e.target_loc
            #i += 1
            #if not S_ <= loc_.zone: # not included
	      #logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      #loc_.zone |= S_
	      #for e_ in self.next(loc_):
		#logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		#if not e_ in  self.waiting: # Can the contrary happen ?
		  #self.waiting.append(e_)
		#else:
		  #logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      #loc_.accepting_zone = loc_.zone & psi[loc_]
	      #if loc_.accepting_zone != loc_.zone:
		#bad = loc_.zone - phi[loc_]
		#loc_.accepting_zone = loc_.accepting_zone.predt(bad)

	      #if not (S_ & loc_.accepting_zone).isEmpty():
		#logger.debug("adding to wait accepting: " + str(loc_))
		#self.waiting.append(e)
            #else:                
	      #logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      #good = loc.accepting_zone
	      #for a in self.nextControllable(loc):
		#good |= a.pred(a.target_loc.accepting_zone)
	      #for a in self.nextUncontrollable(loc):
		#good |= a.pred(a.target_loc.accepting_zone)
	      #bad = loc.zone - phi[loc]
	      #accepting_ = good.predt(bad)
	      #accepting_ &= loc.zone
	      ##print accepting_

	      #if accepting_ != loc.accepting_zone:
		#self.strategy.append(loc)
		#for a in loc.depend:
		  #logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  #e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  #if not e_ in  self.waiting:
		     #self.waiting.append(e_)
		#loc.accepting_zone = accepting_
		#logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      ##loc_.depend.add(t)  
	      #if not t in loc_.depend:
		#loc_.depend.append(t)
	      #logger.debug('Accepting zone:' + str(loc.accepting_zone))
	    
	    #if  self.initial_loc.accepting_zone.hasZero():
	      #solved = True
              #if not all_state: 
		#break

	#logger.info(str(i) + " states visited.")
	
	#if all_state: 
	  #result = {}
	  #for l in self.goal_template.locations:
	    #result[l] = l.accepting_zone
	  #return result
	#else:
	  #return solved

    #def solve_E_weakuntil(self,phi,psi,all_state):
	## all_state: compute and return all the winning states;
	## not all_state: stopped when the initial state is losing; return True or False.
	##TODO: test if the objective is limited to a safety condition w.r.t. some locations.
	## then use a safety game
	
	#init_game()
	#solved = True
	#i = 0
        #while self.waiting:
            #e = self.waiting.pop()
            #t = e.model_transition
            #S = e.source_zone
            #loc = e.source_loc
            #S_ = e.target_zone
            #loc_ = e.target_loc
            #i += 1
            #if not S_ <= loc_.zone: # not included
	      #logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      #loc_.zone |= S_
	      #for e_ in self.next(loc_):
		#logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		#if not e_ in self.waiting: # Can the contrary happen ?
		    #self.waiting.append(e_)
		#else:
		    #logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      #loc_.accepting_zone = loc_.zone - psi[loc_] - phi[loc_]
	      #if loc_.accepting_zone != loc_.zone:
		#good  = psi[loc_]
		#for a in self.nextControllable(loc_):
		  #good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
		#for a in self.nextUncontrollable(loc_):
		  #good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
		#loc_.accepting_zone = loc_.accepting_zone.predt(good-loc_.accepting_zone)
	      
	      #if not (S_ & loc_.accepting_zone).isEmpty():
		#logger.debug("adding to wait accepting: " + str(loc_))
		#self.waiting.append(e)
            #else:                
	      #logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      #bad = loc.accepting_zone
	      #bad += loc.zone.up() - loc.zone # lm comment: S.zone.up() - S.zone is the zone beyond the invariant. In this zone the environnement is winning.
	      ##print "bad: " + str(bad)
	      #good  = psi[loc]
	      #for a in self.nextControllable(loc):
		  #good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      #for a in self.nextUncontrollable(loc):
		  #good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      #accepting_ = bad.predt(good-bad)
	      #accepting_ &= loc.zone
	      ##print accepting_

	      #if accepting_ != loc.accepting_zone:
		#self.strategy.append(loc)
		#for a in loc.depend:
		  #logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  #e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  #if not e_ in self.waiting:
		    #self.waiting.append(e_)
		#loc.accepting_zone = accepting_
		#logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      ##loc_.depend.add(t)  
	      #if not t in loc_.depend:
		#loc_.depend.append(t)
	      #logger.debug('Accepting zone:' + str(loc.accepting_zone))
	          
	    #if self.initial_loc.accepting_zone.hasZero():
              #self.solved = False
	      #if not all_state: 
		#break

	#logger.info(str(i) + " states visited.")
		
	#if all_state: 
	  #result = {}
	  #for l in self.goal_template.locations:
	    #result[l] = l.zone - l.accepting_zone
	  #return result
	#else:
	  #return solved


    #def solve_next(self,phi):
	##TODO: not used for the moment
      
	#init_game()
	#solved = False
        #i = 0
        #while self.waiting:
            #e = self.waiting.pop()
            #t = e.model_transition
            #S = e.source_zone
            #loc = e.source_loc
            #S_ = e.target_zone
            #loc_ = e.target_loc
            #i += 1
            #if not S_ <= loc_.zone: # not included
	      #logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      #loc_.zone |= S_
	      #for e_ in self.next(loc_):
		#logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		#if not e_ in  self.waiting: # Can the contrary happen ?
		  #self.waiting.append(e_)
		#else:
		  #logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      #loc_.accepting_zone = loc_.zone & psi[loc_]
	      #if loc_.accepting_zone != loc_.zone:
		#bad = loc_.zone - phi[loc_]
		#for a in self.nextUncontrollable(loc_):
		  #bad  |= a.pred(a.target_loc.accepting_zone)
		#loc_.accepting_zone = loc_.accepting_zone.predt(bad)

	      #if not (S_ & loc_.accepting_zone).isEmpty():
		#logger.debug("adding to wait accepting: " + str(loc_))
		#self.waiting.append(e)
            #else:                
	      #logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      #good = loc.accepting_zone
	      #for a in self.nextControllable(loc):
		#good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      #bad = loc.zone - phi[loc]
	      #for a in self.nextUncontrollable(loc):
		#bad  |= a.pred(a.target_loc.accepting_zone)
	      #accepting_ = good.predt(bad)
	      #accepting_ &= loc.zone
	      ##print accepting_

	      #if accepting_ != loc.accepting_zone:
		#self.strategy.append(loc)
		#for a in loc.depend:
		  #logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  #e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  #if not e_ in  self.waiting:
		     #self.waiting.append(e_)
		#loc.accepting_zone = accepting_
		#logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      ##loc_.depend.add(t)  
	      #if not t in loc_.depend:
		#loc_.depend.append(t)
	      #logger.debug('Accepting zone:' + str(loc.accepting_zone))
	    
	    #if  self.initial_loc.accepting_zone.hasZero():
	      #solved = True
              #if not all_state: 
		#break

	#logger.info(str(i) + " states visited.")
	
	#if all_state: 
	  #result = {}
	  #for l in self.goal_template.locations:
	    #result[l] = l.accepting_zone
	  #return result
	#else:
	  #return solved
		
    
    def takeTransition(self, zone, model_transition):
        ret_zone = zone.copy()
        ret_zone &= model_transition.guard
        if ret_zone.isEmpty():
            return None
        for (clock, value) in model_transition.update:
	  if value != 0:
            logger.error("Clock reset to non-zero value")
            sys.exit(1)
	  else:
            ret_zone.resetValueInPlace(clock)
        ret_location = model_transition.target
        ret_zone &= ret_location.invariant
	if ret_zone.isEmpty():
            return None

        ret_zone.upInPlace()
        ret_zone &= ret_location.invariant
        ret_zone.extrapolateMaxBoundsInPlace(self.model.clock_bounds)

        transition = DiscreteTransition(zone,model_transition.source,ret_zone,model_transition.target,model_transition)
        
        if not model_transition in ret_location.depend:
	    ret_location.depend.append(model_transition)
        return transition
      
    def next(self, loc):
        ret = []
        for transition in self.goal_template.forward[loc]:       
	  next_discrete_transition = self.takeTransition(loc.zone, transition)
	  if not (next_discrete_transition is None):
	    ret.append(next_discrete_transition)
	self._forward[loc] = ret
        return ret
      
    def nextControllable(self, loc):
      	logger.debug("nextControllable")
        return [t for t in self._forward[loc] if t.model_transition.controllable]
      
    def nextUncontrollable(self, loc):
      	logger.debug("nextUncontrollable")
        return [t for t in self._forward[loc] if not t.model_transition.controllable]
    
    def print_results(self):
      print "Checking formula \'" + self.objective.formula_str + "\' in template \'" + self.goal_template.name + "\'"
      print "=== Winning states ==="
      for loc in self.goal_template.locations:
	print "   Location \'" + loc.name + "\':"
	for state in self.results[loc]:
	  print "    " + self.goal_template.symbolic_state_to_str(state)
      print "\n=== Winning features ==="
      print "  " + self.goal_template.bdd_to_str(self.get_winning_features())
	
      
class SafetyTimedGame(TimedGame):
    def __init__(self,model,goal_template,goal_locations):
	TimedGame.__init__(self,model,goal_template)
	for l in self.goal_template.locations:\
	  l.accepting = (l in goal_locations)
   
    def solve(self):
	init_game()
	solved = True
	i = 0
        while self.waiting:
            e = self.waiting.pop()
            t = e.model_transition
            S = e.source_zone
            loc = e.source_loc
            S_ = e.target_zone
            loc_ = e.target_loc
            i += 1
            if not S_ <= loc_.zone: # not included
	      logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      loc_.zone |= S_
	      if not loc_.accepting:
		for e_ in self.next(loc_):
		  logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		  if not e_ in self.waiting: # Can the contrary happen ?
		      self.waiting.append(e_)
		  else:
		      logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      else:
		loc_.accepting_zone = loc_.zone.copy()
		logger.debug("adding to wait accepting: " + str(loc_))
		self.waiting.append(e)
            else:                
	      logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      bad = loc.accepting_zone
	      bad += loc.zone.up() - loc.zone # lm comment: S.zone.up() - S.zone is the zone beyond the invariant. In this zone the environnement is winning.
	      for a in self.nextUncontrollable(loc):
		  bad  |= a.pred(a.target.accepting_zone)
	      #print "bad: " + str(bad)
	      good  = self.context.getEmptyFederation()
	      for a in self.nextControllable(loc):
		  good |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      accepting_ = bad.predt(good-bad)
	      accepting_ &= S
	      #print accepting_

	      if accepting_ != loc.accepting_zone:
		self.strategy.append(loc)
		for a in loc.depend:
		  logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  if not e_ in self.waiting:
		    self.waiting.append(e_)
		loc.accepting_zone = accepting_
		logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      #loc_.depend.add(t)  
	      if not t in loc_.depend:
		loc_.depend.append(t)
	      logger.debug('Accepting zone:' + str(loc.accepting_zone))
	          
	    if self.initial_loc.accepting_zone.hasZero():
              self.solved = False
	      break

	logger.info(str(i) + " states visited.")
	return solved

    def refine_safety(self,delta_value, delta_name="delta"):
      """ Refine the spoiling strategy to extract parametrically the minimum value of delta.
      """
      self.param_context = pyppl.ParamContext(self.context.clocks,delta_name)
      self.initial_domain = self.param_context.getInitialDomain(0,delta_value)
      logger.debug("initial domain: " + str(self.initial_domain))

      if not hasattr(self,"solved") or self.solved: 
	return self.param_context.getEmptyParamFederation()
	
      for location in self.goal_template.locations:
	location.invariant = pyparamexpr.parse_param_invariant_or_guard(location.invariant_str,delta_name,self.param_context) # overwrite the dbm domain
	location.invariant &= self.initial_domain
	if(location.accepting):
	  location.accepting_param_zone = location.invariant.copy() 
	else:
	  location.accepting_param_zone = self.param_context.getEmptyParamFederation()
      for transition in self.goal_template.transitions:
	  transition.guard = pyparamexpr.parse_param_invariant_or_guard(transition.guard_str,delta_name,self.param_context)  # overwrite the dbm domain
	  transition.update = [self.param_context.getIdVariable(clock.name.split('.')[1]) for (clock,value) in transition.update] # overwrite the dbm clocks update
		
      self.waiting = self.strategy
              
      i = 0
      while self.waiting:
	current = self.waiting.pop(0)
	if current.accepting_zone.isEmpty():
	  continue

	logger.debug("Refining: " + str(current))
	i += 1
  
        bad = ~current.invariant
        bad &= self.initial_domain
        for transition in self.nextUncontrollable(current):
	  if(not transition.target_loc.accepting_param_zone.isEmpty()):
	    bad  |= transition.target_loc.accepting_param_zone.pred(transition.model_transition.guard & current.invariant,transition.model_transition.update)
	good  = self.param_context.getEmptyParamFederation()
	for transition in self.nextControllable(current):
	  pred_acc = transition.target_loc.invariant - transition.target_loc.accepting_param_zone
	  good |= pred_acc.pred(transition.model_transition.guard & current.invariant,transition.model_transition.update) 
	good -= bad
	good &= self.initial_domain
	
	current.accepting_param_zone = bad.predt(good)
	current.accepting_param_zone &= current.invariant
	#accepting_.pairwiseReduce()
	
	logger.debug(str(i) + " states refined;")
	logger.debug("Refined location: " + str(current) + " ; accepting domain: " + str(current.accepting_param_zone))
	if current == self.initial_loc:
	  refine_values = self.initial_loc.accepting_param_zone.hasZeroParamDomain()
	  if not refine_values.isEmpty():
	    logger.info(str(i) + " states refined;")
	    return refine_values.minimizeParam()
      logger.error("no winning value in parameter domain: " + str(self.initial_loc.accepting_param_zone))
      return None	
    

class ReachabilityTimedGame(TimedGame):
    def __init__(self,model,goal_template,goal_locations):
	TimedGame.__init__(self,model,goal_template)
	for l in self.goal_template.locations:
	  l.accepting = (l in goal_locations)
	     
    def solve_until(self):
	init_game()
	solved = False
        i = 0
        while self.waiting:
            e = self.waiting.pop()
            t = e.model_transition
            S = e.source_zone
            loc = e.source_loc
            S_ = e.target_zone
            loc_ = e.target_loc
            i += 1
            if not S_ <= loc_.zone: # not included
	      logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
	      loc_.zone |= S_
      	      if not loc_.accepting:
		for e_ in self.next(loc_):
		  logger.debug("adding to wait next: " + str(e_.target_loc) + "#" + str(e_.target_zone))
		  if not e_ in  self.waiting: # Can the contrary happen ?
		    self.waiting.append(e_)
		  else:
		    logger.debug("already in wait: " + str(e_.target_loc) + "#" + str(e_.target_zone))
	      else:
		loc_.accepting_zone = loc_.zone
		logger.debug("adding to wait accepting: " + str(loc_))
		self.waiting.append(e)
            else:               
	      logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
	      good = loc.accepting_zone
	      for a in self.nextControllable(loc):
		good |= a.pred(a.target_loc.accepting_zone)
	      bad = self.context.getEmptyFederation()
	      for a in self.nextUncontrollable(loc):
		bad  |= a.pred(a.target_zone - a.target_loc.accepting_zone)
	      accepting_ = good.predt(bad)
	      accepting_ &= loc.zone
	      #print accepting_

	      if accepting_ != loc.accepting_zone:
		self.strategy.append(loc)
		for a in loc.depend:
		  logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
		  e_ = DiscreteTransition(a.source.zone,a.source,a.target.zone,a.target,a)
		  if not e_ in  self.waiting:
		     self.waiting.append(e_)
		loc.accepting_zone = accepting_
		logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
	      #loc_.depend.add(t)  
	      if not t in loc_.depend:
		loc_.depend.append(t)
	      logger.debug('Accepting zone:' + str(loc.accepting_zone))
	    
	    if  self.initial_loc.accepting_zone.hasZero():
	      solved = True
	      break

	logger.info(str(i) + " states visited.")
	
	return solved
        
#if __name__ == '__main__':

    #model = pyuppaal.NTA.from_xml(sys.argv[1])
    
    #dbmpyuppaal.parse_nta_dbm(model)
    
    #goal_template = model.getTemplateByName(sys.argv[2].split('.')[0])
    #goal_location_name = sys.argv[2].split('.')[1]
    
    #dbmpyuppaal.parse_templates_dbm(model,[goal_template])

    #game = SafetyTimedGame(model, goal_template, goal_location_name) 
    #print game.solve()
    #game.printDot('game.dot')

    
